/* 
 * 
 * Copyright (c) Year(s), 2013, Luis M. S. Russo and Alexandre 
 * P. Francisco / KDBIO / INESC-ID, <qhv@kdbio.inesc-id.pt> 
 *  
 * Any published media that is related with to use of the distributed 
 * software, or derived software, must contain a reference to "Extending 
 * quick hypervolume. Luís M. S. Russo, Alexandre P. Francisco: 
 * J. Heuristics 22(3): 245-271 (2016)".
 *  
 * Permission to use, copy, modify, and/or distribute this software for 
 * any purpose with or without fee is hereby granted, provided that the 
 * above copyright notice and this permission notice appear in all 
 * copies. 
 *  
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL 
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL 
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR 
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR 
 * PERFORMANCE OF THIS SOFTWARE. 
 * 
 */

#include <stdlib.h>
#include <pthread.h>
#include <stdio.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>	/* basic system data types */
#include <limits.h>		/* PIPE_BUF */
#include <string.h>

#include "tpool.h"

/* This is the maximal number of cores in PC */
#ifndef CORES
#define CORES 64
#endif /* CORES */

/* Number of dimensions import from existing code */

/* Maximum number of simultaneous jobs */
#define SIMJs (CORES+1)
/* Maximum number of simultaneous tasks */
/* The last SIMJs, guarantees a task per job. */
#define SIMTs (4*SIMJs)

/// Structure that stores processable information of a task
struct task {
  /// Pass job Id to thread
  toolBox* b;                   /**< When the task is assigned to a worker it gets a toolBox. */
  int jid;                      /**< Job id. What problem is being solved. */
  taskData d;
};

typedef struct task* task;

/// Possible queue states
enum Qstate  {active, stopped};

/// Pointers to functions
struct fns{
  void  (*run)(toolBox* b, taskData* d, const jobData* jd);
  void  (*openBox)(toolBox* b);
  void  (*closeBox)(toolBox* b);
  int   (*priority)(taskData* d);     /**< Priority for task. */
  void  (*jobCopy)(jobData* des, jobData* src);
  void  (*taskCopy)(taskData* dst, taskData* src);
  void  (*merge)(jobData* dst, taskData* src);
  void  (*publish)(int jid, jobData* jd);
};

/// SIMJs is queue size
struct jobQueue{
  pthread_mutex_t* mutex;       /**< Exclusive read or write operations to tasks and jobs */
  pthread_cond_t cond;          /**< Use to warn console of available space. */
  int head;                     /**< Position where you can insert a new job. */
  int tail;                     /**< Position of the last job being processed. */
  int C[SIMJs];                 /**< Array that counts the number tasks on a job. */
  jobData R[SIMJs];             /**< Array with results. */  
};

struct taskQueue {
  pthread_mutex_t mutex;        /**< Exclusive read or write operations to tasks and jobs */
  pthread_cond_t cond;          /**< Use to warn workers after posting. */
  pthread_cond_t mem;           /**< Use to warn priority postings of available space. */
  int waiters;                  /**< How many workers waiting on first cond */
  int minGrade;                 /**< Only tasks that score higher are admissable. */
  enum Qstate s;                /**< If all tasks ended, empty and stopped it finishes */
  /// The array is organized in the following order:
  int memDrop;                  /**< Position to drop task structure */
  int memGet;                   /**< Position to grab a task structure */
  int head;                     /**< Position where you can insert a new task. */
  int tail;                     /**< Position of popable element. */
  int heavyTail;                /**< Number of elements behind tail, that are being processed. */
  task A[SIMTs];                /**< The array. */
};

static struct fns F;
static struct taskQueue Q = {
  PTHREAD_MUTEX_INITIALIZER,
  PTHREAD_COND_INITIALIZER,
  PTHREAD_COND_INITIALIZER
};
static struct jobQueue Jq = {
  &(Q.mutex),
  PTHREAD_COND_INITIALIZER
};

/* An array of tool boxes, one for each worker. */
static toolBox toodles[CORES];

/* static void printState(void); */

/// Only functions that do not assume mutex

/** 
 * 
 * @return True iff workers are still needed.
 */
static int activeQ(void);

/* #define  pthread_mutex_lock(A)  pthread_mutex_lock(A); printf("mutex Locked\n"); printState(); */
/* #define  pthread_mutex_unlock(A)  pthread_mutex_unlock(A); printf("mutex UnLocked\n") */

/** 
 *
 * Safely finish the queue
 * 
 */
static void halt(void)
{
  int actv;

  pthread_mutex_lock(&Q.mutex);
  Q.s = stopped;
  actv = !activeQ();
  pthread_mutex_unlock(&Q.mutex);

  if(actv)
    pthread_cond_broadcast(&Q.cond); /* wake sleeping threads. */
}

/// These functions assume you have the queue mutex

/** 
 * 
 * @return Resouces in quartiles, 0 to 4.
 */
static int resources(void)
{
  return ((Q.memGet - Q.heavyTail)*4)/SIMTs;
}

/** 
 * 
 * @return Number of pending tasks.
 */
static int tasksN(void)
{
  return Q.head - Q.tail;
}

/** 
 * 
 * @return Number of pending jobs.
 */
static int jobsN(void)
{
  return Jq.head - Jq.tail;
}

/** 
 * 
 * @return True iff workers are still needed.
 */
static int activeQ(void)
{
  return (active == Q.s) || (Jq.tail < Jq.head);
}

/* static void printState(void) */
/* { */
/*   printf("Q info: %d jobs, %d tasks, %d waiters\n", jobsN(), tasksN(), Q.waiters ); */
/* } */

/** 
 * Assumes you know there is a task to process. Undefined behaviour otherwise.
 *  
 * @return The next task that needs processing
 */
static task pop(void)
{
  task t;

  assert(tasksN() > 0);

  t = Q.A[Q.tail % SIMTs];
  Q.A[Q.tail % SIMTs] = NULL;
  Q.tail++;

  return t;
}

/** 
 * Submit a value back to the job queue
 * 
 * @param jid The job id
 * @param r The value
 *  
 * @return Signal that job space is available
 */
static int submit(int jid, taskData* d)
{
  int signal;

  /// Otherwise there is queue corruption.
  assert(Jq.tail <= jid && jid <= Jq.head);

  (*F.merge)(&(Jq.R[jid % SIMJs]), d);
  /* Jq.R[jid % SIMJs]+=r; */
  Jq.C[jid % SIMJs]--;          /**< One less task for this job */

  signal = (jobsN() == SIMJs && 0 == Jq.C[Jq.tail % SIMJs]);

  while(jobsN() > 0 && 0 == Jq.C[Jq.tail % SIMJs])
    {                           /**< Dump computed results. */
      (*F.publish)(Jq.tail, &(Jq.R[Jq.tail % SIMJs]));
      /* Do not clean up, jobs are initialized */
      Jq.tail++;
    }

  return signal;
}

/** 
 * Returns the task back to the Queue. Sinals are delayed for efficiency.
 * 
 * @param t The task to recycle
 *
 * @return Signal that task space is available
 */
static int recycle(task t)
{
  int signal;

  signal = (resources() == 4);  /**< There was a resource shortage */

  Q.A[Q.memDrop % SIMTs] = t;
  Q.memDrop++;
  Q.heavyTail++;                 /**< One less task to do */

  assert(Q.heavyTail <= Q.tail);

  return signal;
}

/** 
 * Asks for a task container
 * 
 * @param p A priority parameter. Pass gigantic number for VIPs.
 * 
 * @return NULL on no space and task pointer otherwise.
 */
static task queryTs(int p)
{
  task t;
  
  t = NULL;

  /* printf("Admission criteria: Q.minGrade %d < p %d\n", Q.minGrade, p); */

  if(Q.minGrade < p)
    {    /* Hurray, you passed task admission */
      Q.minGrade += resources();   /* Complain about lack of resources */
      if(resources() < 4)         
        { /* Huff, you still got the resource. */
          t = Q.A[Q.memGet % SIMTs];
          Q.memGet++;

          if(Q.memGet > Q.memDrop) /* But need to alloc it. */
            {
              Q.memDrop++;
              t = (task) calloc(1, sizeof(struct task));
            }

          assert(Q.memGet <= Q.memDrop);
        }
    }
  else /* We regret to inform you ... bla bla ... no task admission. */
    Q.minGrade += resources() - 3;   /* Complain about waisted resources */

  return t;
}

/** 
 *
 * Posts i to the queue, if possible.
 * If the tid is 0 then it locks until it gets the post.
 *
 * IMPORTANT: Automatically frees mutex.
 *
 * @param pri If it is 0 give max priority
 * @param i   The task to upload to the queue
 *
 * @return True if upload successful.
 */
static int post(int pri, int jid, taskData* d)
{
  int r;
  task t;
  int signal;
  int p;                        /**< Priority */

  /* printf("Trying to post task\n"); */
  /* printState(); */

  signal = 0;

  p = (*F.priority)(d);
  if(0 == pri)
    p = INT_MAX;
  t = queryTs(p);

  if(0 == pri)
    {
      while(t == NULL)
        { /* LOOP until there is room to post task */
          pthread_cond_wait(&Q.mem, &Q.mutex);
          t = queryTs(p);
        }
    }

  r = (t != NULL);
  if(r)
    {                   /**< Copy memory */
      /* printf("Success post\n"); */
      t->jid = jid;
      Jq.C[t->jid % SIMJs]++;         /**< Update task count before returning mutex */
      pthread_mutex_unlock(&Q.mutex); /**< Give mutex back, the copy will take a while. */
      (*F.taskCopy)(&(t->d), d);
      t->b = NULL;              /**< A new task has no toolBox. */

      pthread_mutex_lock(&Q.mutex);
      signal = (activeQ() && Q.waiters > 0); /**< There may be sleeping threads */
      Q.A[Q.head % SIMTs] = t;      /**< put on queue. */        
      Q.head++;      

      assert(Q.head <= Q.memGet);
      assert(Q.memGet <= Q.memDrop);
    }
  pthread_mutex_unlock(&Q.mutex); /* Also releases by default */

  if(signal)
    pthread_cond_signal(&Q.cond); /* wake sleeping threads. */

  return r;
}

/** 
 *
 * Starts a new job, by uploading the first task.
 *
 * @param i The task to upload to the queue.
 */
void postJob(jobData* jd, taskData* d)
{
  int jid;

  pthread_mutex_lock(&Q.mutex);

  while(jobsN() == SIMJs) /* Job queue is full */
    pthread_cond_wait(&Jq.cond, Jq.mutex);

  jid = Jq.head;             /**< You need to fill this before posting */
  Jq.head++;                 /* Clean space is available, just update counter */

  (*F.jobCopy)(&(Jq.R[jid % SIMJs]), jd);

  post(0, jid, d);            /**< Post the first task */
}

void prun(taskData* d, int p)
{
  task t;

  t = (task) container_of(d, struct task, d); /* The d's mean diferent things. */

  /* Assuming taskData actually is in a task */
  if(p)
    { /* Try to Parallelize the computation */
      pthread_mutex_lock(&Q.mutex);
      if(!post(1, t->jid, d)) /* Not crucial, use task priority */
        (*(F.run))(t->b, d, &(Jq.R[t->jid % SIMJs]));  /**< Do it yourself */
    }
  else
    (*(F.run))(t->b, d, &(Jq.R[t->jid % SIMJs]));  /**< Do it yourself */
}

void* worker(void *threadid)
{
  task t;
  int tid;
  /// Use to signal events
  int signalMem;
  int signalJob;
  int signalDismiss;
  toolBox* b;

  tid = *((int*)threadid);

  b = &(toodles[tid-1]);
  (*F.openBox)(b);             /**< Start tools */

  do
    {
      signalJob = 0;
      signalMem = 0;
      signalDismiss = 0;

      pthread_mutex_lock(&Q.mutex);
      while(activeQ() && 0 == tasksN())      /**< Wait until there is something to do */
        {
          Q.waiters++;
          pthread_cond_wait(&Q.cond, &Q.mutex);
          Q.waiters--;
          /* printf("Got a Job\n"); */
          /* printState(); */
        }

      if(activeQ())
        { /* Otherwise stop worker */
          t = pop();                /**< Get the next task */
          pthread_mutex_unlock(&Q.mutex);
          t->b = b;
          (*F.run)(t->b, &(t->d), &(Jq.R[t->jid % SIMJs]));  /**< Go run your task */

          pthread_mutex_lock(&Q.mutex);
          signalJob = submit(t->jid, &(t->d));  /**< Submit result to the job queue  */
          signalMem = recycle(t);               /**< Return task structure to task queue */
          signalDismiss = !activeQ();
        }
      pthread_mutex_unlock(&Q.mutex);

      /* Signal events */
      if(signalJob)
        pthread_cond_signal(&Jq.cond); /* wake sleeping threads. */        
      if(signalMem)
        pthread_cond_signal(&Q.mem);   /* wake sleeping threads. */        
      if(signalDismiss)
        pthread_cond_broadcast(&Q.cond); /* wake sleeping threads. */
    }
  while(activeQ());

  (*F.closeBox)(b);             /**< Close tools */

  pthread_exit(NULL);
}

void launch(
            int w,
            void* (*console)(void *threadid),
            void  (*run)(toolBox* b, taskData* d, const jobData* jd), 
            void  (*openBox)(toolBox* b),
            void  (*closeBox)(toolBox* b),
            int   (*priority)(taskData* d),
            void  (*jobCopy)(jobData* des, jobData* src),
            void  (*taskCopy)(taskData* dst, taskData* src),
            void  (*merge)(jobData* dst, taskData* src),
            void  (*publish)(int jid, jobData* jd)
            )
{
  int i;
  pthread_t wths[CORES];        /**< Worker threads */
  pthread_t cthread;            /**< Console thread */
  int tids[CORES];

  F.run = run;
  F.openBox = openBox;
  F.closeBox = closeBox;
  F.priority = priority;
  F.jobCopy = jobCopy;
  F.taskCopy = taskCopy;
  F.merge = merge;
  F.publish = publish;

  if(CORES < w)
    w = CORES;
  if(1 > w)
    w = 1;

  /* start all the worker threads */
  printf("Using %d threads\n", w);
  /* toodles = (toolBox*) calloc(w, sizeof(toolBox)); */
  memset(toodles, 0x0, w*sizeof(toolBox));

  i = 0;
  while(i < w)
    {
      tids[i] = i+1;
      pthread_create(&wths[i], NULL, worker, &tids[i]);
      i++;
    }

  /* Create console only after you have the workers */

  pthread_create(&cthread, NULL, console, NULL);
  pthread_join(cthread, NULL);  
  halt();

  /* Afther console finishes wait for workers. Forced to finish by halt. */

  i = 0;
  while(i < w)
    {
      pthread_join(wths[i], NULL);
      i++;
    }
  
  i = 0;
  while(i < SIMTs)
    {
      if(NULL != Q.A[i])
        free(Q.A[i]);
      i++;
    }

  /* free(toodles); */
}
